/* --------------------------------------------------------------- */
/*                       MHDIFF                                    */
/*                                                                 */
/*          Multiplicative Heterogeneous DIFFusion                 */
/*                                                                 */
/* --------------------------------------------------------------- */
/*                                                                 */
/*    Version 2.1     B                                            */
/*                                                                 */
/*    (C)  1995     David Strang                                   */
/*                                                                 */
/*    last modified: October 1995                                  */
/*                   November 1995                                 */
/*                   June 1996                                     */
/*                   January 2001                                  */
/*                   November 2001                                 */
/*                                                                 */
/*    For program description and for citation in published        */
/*    work using this software, see David Strang,                  */
/*    "MHDIFF: User Documentation," Technical Report 95-3,         */
/*    Department of Sociology, Cornell University.)                */
/*                                                                 */
/*   ------------------------------------------------------------- */



%macro mhdiffb2; 

proc iml workspace=&wspace;
reset nocenter pagesize=72;

verbose = {&verbose};
if verbose = 1 then
   reset name;
else
   reset noname;


/* spell constructs */

st = {&st}; 
et = {&et}; 
ss = {&ss}; 
es = {&es}; 

use_et2 = {&use_et2};
if (use_et2 = 1) then do;
   st2 = {&st2};
   et2 = {&et2};
   ss2 = {&ss2};
   es2 = {&es2};
end;

save_all = {&save_all};

id = {&id}; 
mvalue = {&mvalue};

/* covariates */

aterm = {&aterm};              
/* if aterm ^= 0 then  */
      avars = {&avars};

bterm = {&bterm};              
/* if bterm ^= 0 then */
      bvars = {&bvars};

cterm = {&cterm};              
/* if cterm ^= 0 then */
      cvars = {&cvars};

dterm = {&dterm};              
if (dterm ^= 0) then do; 
    dopts = {&dopts};
    dvecs = {&dvecs};
    dnames = {&dnames};
    do d = 1 to nrow(dopts);
       if ((dopts[d,1] = 3) & (dopts[d,4] = 2)) then tiesize = {&tiesize};
    end;       
end; 
else do;
    dopts = {0 0 0 0 1};
    dvecs = {-1};
    dnames = {"d-fault"};
end;

/* options */


sametime = {&sametime}; 

simlapse = {&simlapse};

insum = {&insum};   

igroup = {&igroup}; 
   
data_spl = {&sdata};



/* obsolete keywords */

/* wrtvars = {&wrtvars}; */



/* ----------- end of command input --------------------------------- */


/* ----------- read data -------------------- */

if (&indir = 0) then
    reset deflib = 'WORK';
else
    reset deflib = 'difflib'; 

sort data=&sdata out=data_s by &sort_var;  
      /* if std diffusion using et, sort_var is the name of the et variable */
      /* if diffusion using et2, sort_var is the name of the et2 variable */

use data_s;
read all into x[COLNAME=xvars];
xvars = xvars`;
close data_s;


%do ijk = 1 %to &dterm;
   if dopts[&ijk,1] = 2 then do;
       use nxnd_%eval(&ijk);
       read all into nxn_%eval(&ijk);
       close nxnd_%eval(&ijk);
       nxnid_%eval(&ijk) = nxn_%eval(&ijk)[,1];
       if (dopts[&ijk,4] ^= 0) then do;
           v = nxn_%eval(&ijk)[,1];
           nxn_%eval(&ijk) = nxn_%eval(&ijk)[,3:ncol(nxn_%eval(&ijk))];
       end;
       else 
           nxn_%eval(&ijk) = nxn_%eval(&ijk)[,2:ncol(nxn_%eval(&ijk))];
       nn = nxn_%eval(&ijk); 
       if (dopts[&ijk,3] > 0) then do;
          nn = nxn_%eval(&ijk); 
          if (dopts[&ijk,3] = 1) then
              nxn_%eval(&ijk) = nxn_%eval(&ijk)[<>,<>] - nxn_%eval(&ijk);
          else if (dopts[&ijk,2] = 1) then do;
              maxd_i = nxn_%eval(&ijk)[,<>];
              maxd_i = repeat(maxd_i,1,ncol(nn));
              sumd_i = nxn_%eval(&ijk)[,+];
              sumd_i = repeat(sumd_i,1,ncol(nn));
          end;
          else if (dopts[&ijk,2] = 2) then do;
              maxd_i = nxn_%eval(&ijk)[<>,];
              maxd_i = repeat(maxd_i,nrow(nn),1);
              sumd_i = nxn_%eval(&ijk)[+,];
              sumd_i = repeat(sumd_i,nrow(nn),1);
          end;
          if (dopts[&ijk,3] = 2) then
               nxn_%eval(&ijk) = maxd_i - nxn_%eval(&ijk);
          else if (dopts[&ijk,3] = 3) then
               nxn_%eval(&ijk) = nxn_%eval(&ijk)/sumd_i;
          else if (dopts[&ijk,3] = 4) then
               nxn_%eval(&ijk) = (maxd_i - nxn_%eval(&ijk))/sumd_i;
          else if (dopts[&ijk,3] = 5) then    
               nxn_%eval(&ijk) = (maxd_i - nxn_%eval(&ijk))/
                                 (((nrow(nn) - 1) * maxd_i) - sumd_i);
          else if (dopts[&ijk,3] = 6) then do;
               v = repeat(2,nrow(nn),1);
               vmx = repeat(v,1,ncol(nn));
               nxn_v1 = (maxd_i - nxn_%eval(&ijk))##vmx;
               nxn_v2 = (((nrow(nn) - 1) * maxd_i) - sumd_i)##vmx;
               nxn_%eval(&ijk) = nxn_v1/nxn_v2;
          end;

       end;
   end;
%end;


if (aterm = 1) then 
   covnames = "a_0";
else if (aterm = 2) then
   covnames = concat("a: ",xvars[avars]);
else if (aterm = 3) then
   covnames = "a_0" // concat("a: ",xvars[avars]);

dif_strt = nrow(covnames) + 1;
if bterm = 1 then
   covnames = covnames // "b_0";
else if (bterm = 2) then
   covnames = covnames // concat("b: ",xvars[bvars]);
else if bterm = 3 then
   covnames = covnames // "b_0" // concat("b: ",xvars[bvars]);
if cterm ^= 0 then
   covnames = covnames // concat("c: ", xvars[cvars]);
d_strt = nrow(covnames) + 1;
if dterm ^= 0 then do;
   dnamel = dnames[1:dterm];
   covnames = covnames // concat("d: ",dnamel);
end;
if (aterm = 1) | (aterm = 3) then 
   cov_nint = covnames[2:nrow(covnames)];
else
   cov_nint = covnames;


print "Spell data : " data_spl;

print "Model covariates are: ";
print covnames;

print "Model specifications are: ";

if aterm ^= 0 then print "ATERM " aterm "AVARS " avars;

if bterm ^= 0 then print "BTERM " bterm "BVARS " bvars;

if cterm ^= 0 then print "CTERM " cterm "CVARS " cvars;

if dterm ^= 0 then print "DTERM " dterm,
                         " DOPTS" dopts,
                         " DVECS" dvecs,
                         " DNAMES" dnamel;

print "SAMETIME " sametime;
if (sametime = 1) then print "SIMLAPSE " simlapse;

print "INSUM    " insum;
if (insum = 1) then print "IGROUP   " igroup;


/* -----  descriptive statistics --- */

start dstats(amx,rnames);
    xmean = amx[:,];
    xmax =  amx[<>,];
    xmin =  amx[><,];

    xmean_r = repeat(xmean,nrow(amx),1);
    xdiff = amx - xmean_r;
    xsd2 = xdiff##2;
    xvar = xsd2[+,]/(nrow(amx) - 1);
    xxsd = sqrt(xvar);

    xdesc = xmean // xxsd // xvar // xmin // xmax;
    xdesc = xdesc`;
    desc_col = {"Mean" "S.D." "Var" "Minimum" "Maximum"};

    print xdesc [r=rnames c=desc_col format = 10.4];
    return;
finish dstats;

start mxdim(amx,mname);
    nrows_ = nrow(amx);
    ncols_ = ncol(amx);
    print  mname " has N Cases (rows) " nrows_  
                 " and N Variables (columns) " ncols_;
    return;
finish mxdim;

/* display original spell data */

if verbose then do;
    run mxdim(x,data_spl);
    run dstats(x,xvars);

    ts_x = x[,id] || x[,st] || x[,et] || x[,es];
    if (aterm = 2) | (aterm = 3) then ts_x = ts_x || x[,avars];
    tsrows = min(120,nrow(ts_x));
    ts_x = ts_x[1:tsrows,];
    print "Original Spell Data (first 120 cases)" data_spl ts_x;
end;

/* display events that will be modeled */

diff_evt = x[loc(x[,es] ^= x[,ss]),];

xnevents = nrow(diff_evt);
xtimet = sum(x[,et] - x[,st]);
print "Number of Modeled Events " xnevents;
print "Total Time at Risk of Modeled Event " xtimet;

if verbose then do;
    call mxdim(diff_evt,"diff_evt"); 
    ts_evt = diff_evt[,id] || diff_evt[,st] || diff_evt[,et] || 
             diff_evt[,ss] || diff_evt[,es];
    print "Modeled Event Times" ts_evt;
end;

/* --------- define event (influencer) matrix ------------- */

if (bterm ^= 0) | (cterm ^= 0) | (dterm ^= 0) then do;

dcovname = covnames[dif_strt:nrow(covnames)];
   
/* note: x must be sorted by ending time of influencers */
/* by et if using es, by et2 if using es2 */

evt = x[loc(x[,es] ^= x[,ss]),];

if (use_et2 = 1) then do;
   evt = x[loc(x[,es2] ^= x[,ss2]),];
   evt[,st] = evt[,st2];
   evt[,et] = evt[,et2];
   evt[,ss] = evt[,ss2];
   evt[,es] = evt[,es2];
end;

/* display influencing events */
    inevents = nrow(evt);
    print "Number of Influencing Events " inevents;

if verbose then do;
    call mxdim(evt,"evt"); 
    its_evt = evt[,id] || evt[,st] || evt[,et] || evt[,ss] || evt[,es];
    print "Influencing Event Times" its_evt;
end;



/*  x -> diffx */

if (sametime = 0) | (simlapse = 0) then do;
do i=1 to nrow(x);
   if (insum = 1) & any(evt[,igroup] = x[i,igroup]) then 
      evt_i = evt[loc(evt[,igroup] = x[i,igroup]),];
   else
      evt_i = evt;
   j = 0;
   done_i = 0;
   if (insum = 1) & ^any(evt[,igroup] = x[i,igroup]) then do;
      diffx = diffx // x[i,];
      done_i = 1;
   end;
   do while (done_i = 0);
      j = j + 1;
      if (x[i,st] < evt_i[j,et]) & (x[i,et] > evt_i[j,et]) then do;
         diffx = diffx // x[i,];
         diffx[nrow(diffx),et] = evt_i[j,et];
         diffx[nrow(diffx),es] = x[i,ss];
         x[i,st] = evt_i[j,et];
      end;
      if (x[i,et] <= evt_i[j,et]) | (j = nrow(evt_i)) then do;
         diffx = diffx // x[i,];
         done_i = 1;
      end;
   end;
end;
end;
else if (sametime = 1) then do;
do i=1 to nrow(x);
   if (insum = 1) & any(evt[,igroup] = x[i,igroup]) then 
      evt_i = evt[loc(evt[,igroup] = x[i,igroup]),];
   else
      evt_i = evt;
   j = 0;
   done_i = 0;
   if (insum = 1) & ^any(evt[,igroup] = x[i,igroup]) then do;
      diffx = diffx // x[i,];
      done_i = 1;
   end;
   do while (done_i = 0);
      j = j + 1;
      if (x[i,st] < evt_i[j,et]) & (x[i,et] >= evt_i[j,et]) then do;
         diffx = diffx // x[i,];
         diffx[nrow(diffx),et] = evt_i[j,et] - simlapse;
         diffx[nrow(diffx),es] = x[i,ss];
         x[i,st] = evt_i[j,et] - simlapse;
     end;
     if (x[i,et] <= evt_i[j,et]) | (j = nrow(evt_i)) then do;
        diffx = diffx // x[i,];
        done_i = 1;
     end;
   end;
end;
end;




diffx = diffx[loc(diffx[,st] ^= diffx[,et]),];

if verbose then do;
   run dstats(diffx,xvars);
   call mxdim(diffx,"diffx");
   ts_dx = diffx[,id] || diffx[,st] || diffx[,et] || diffx[,es];
   if (aterm = 2) | (aterm = 3) then ts_dx = ts_dx || diffx[,avars[1]];
   tsrows = min(120,nrow(ts_dx));
   ts_dx = ts_dx[1:tsrows,];
   print "Constructed Diffusion Spell Data (first 120 spells) ";
   print ts_dx;
end;

/* free x; */

/* -------    generate diffusion covariates   --------    */

if (dterm ^= 0) then do;
      if any(dopts[,2] = 4) | any(dopts[,2] = 5) then do;
          nevts = repeat(0,nrow(diffx),1);
          max_d = repeat(0,nrow(dopts),1);
          max_dj = repeat(0,nrow(diffx),1);
      end;
end;


do i=1 to nrow(diffx); 
   
   if insum = 1 then 
      inspell = ((evt[,et] > diffx[i,st]) & (evt[,et] < diffx[i,et])
                  & (evt[,igroup] = diffx[i,igroup]));
   else	
      inspell = ((evt[,et] > diffx[i,st]) & (evt[,et] < diffx[i,et]));


   if inspell = 1 then do;
   icase = diffx[i,id] || diffx[i,st] || diffx[i,et] || diffx[i,es];
      print "Warning: event located within spell in diffx" i icase  inspell;
   end;

   if (sametime = 0) then
      evt_i = evt[,et] < diffx[i,et];
   else if (sametime = 1) then do;
      evt_i = evt[,et] <= diffx[i,et];
      evt_i = evt_i # (evt[,id] ^= diffx[i,id]);
   end;
   if insum = 1 then
      evt_i = evt_i # (evt[,igroup] = diffx[i,igroup]);


  if (bterm ^= 0) then do;
       if (bterm = 1) then
           bline = evt_i;
       else if (bterm = 2) then
           bline = diffx[i,bvars] @ evt_i;
       else if (bterm = 3) then
           bline = (1 || diffx[i,bvars]) @ evt_i;
       covb_s = bline[+,];
   end;

   if (cterm ^= 0) then do;
       covc = evt[,cvars] # evt_i;
       covc_s = covc[+,];
   end;

   if (dterm ^= 0) then do;
     %do d = 1 %to &dterm;
       dvars = dvecs[&d,1:dopts[&d,5]];  
       if (dopts[&d,1] = 0) then do;
          dpart = repeat(0,nrow(evt),ncol(dvars));
          if (dopts[&d,3] ^= 0) then do;
             dpart =
     (evt[,dvars] - repeat(diffx[i,dvars],nrow(evt),1))##dopts[&d,3];
             dvec = dpart[,+];
             if (dopts[&d,4] ^= .) then
                 dvec = dvec # (dvec > dopts[&d,4]);
          end;
          else if (dopts[&d,3] = 0) then do;
             dpart = evt[,dvars] - repeat(diffx[i,dvars],nrow(evt),1);
             dvec = dpart[,+];
             if (dopts[&d,4] ^= .) then 
                 dvec = (dvec > dopts[&d,4]);
             else do;
                 dpos = (dvec > 0);
                 dneg = (dvec < 0);
                 dvec = dpos - dneg;
             end;
          end;
          dline = evt_i` * dvec;
       end;
       else if (dopts[&d,1] = 1) then do;
          dpart = repeat(0,nrow(evt),ncol(dvars));
          if (dopts[&d,3] ^= 0) then do;
             dpart =
     abs(evt[,dvars] - repeat(diffx[i,dvars],nrow(evt),1))##dopts[&d,3];
             if (dopts[&d,4] ^= 0) then
                 dvec = (dpart[,+])##(1/dopts[&d,4]);
             else
                 dvec = dpart[,<>];
          end;
          else if (dopts[&d,3] = 0) then do;
             dpart =
               evt[,dvars] = repeat(diffx[i,dvars],nrow(evt),1); /* eq */
             if (dopts[&d,4] = 0) then
                dvec = dpart[,<>];
             else if (dopts[&d,4] ^= 0) then
                dvec = dpart[,+];
          end;
          if (dopts[&d,2] = 2) | (dopts[&d,2] = 3) then 
              dvec = abs((dvec ^= 0) - 1);
          if (dopts[&d,2] = 4) then do;
              dvec_i = dvec # evt_i;
              max_dj[i] = max(dvec_i); 
              if (max_dj[i] > max_d[&d]) then max_d[&d] = max_dj[i];
              nevts[i] = evt_i[+];
          end;
          dline = evt_i` * dvec;
          if (dopts[&d,2] = 3) & (sum(evt_i) ^= 0) then 
              dline = dline/sum(evt_i);
       end;
       else if (dopts[&d,1] = 2) then do;     /* nxn proximity data */
          nxn = nxn_%eval(&d);
          nxnids = nxnid_%eval(&d);
          evtids_i = evt[,id] # evt_i;
          if any(evtids_i) & any(nxnids = diffx[i,id]) then do;
               evtids_i = evtids_i[loc(evtids_i ^= 0)];
               pij = repeat(0,1,nrow(evtids_i));
               do ij = 1 to nrow(evtids_i);
                  if any(evtids_i[ij] = nxnids) then do;
                     if (dopts[&d,2] = 1) then 
    pij[ij] = nxn[loc(nxnids = diffx[i,id]),loc(nxnids = evtids_i[ij])];
                     else if (dopts[&d,2] = 2) then
    pij[ij] = nxn[loc(nxnids = evtids_i[ij]),loc(nxnids = diffx[i,id])];
                  end;
               end;
               dline = pij[+];    
          end;
          else do;
                dline = 0; 
          end;
       end;
       else if (dopts[&d,1] = 3) then do;   /* pairs data */
          if (dopts[&d,2] = 1) then do;        /* case has who influences */
              dvec  = (evt[,dvars] = diffx[i,dopts[&d,3]]); 
              if (dopts[&d,4] = 2) then 
                  dvec = dvec # evt[,tiesize[&d,1:dopts[&d,5]]];
          end;
          else if (dopts[&d,2] = 2) then do;   /* case has who infl by */
               dvec = (repeat(evt[,dopts[&d,3]],1,ncol(dvars)) = 
                         repeat(diffx[i,dvars],nrow(evt),1));
               if (dopts[&d,4] = 2) then 
                   dvec = dvec # 
                       repeat(diffx[i,tiesize[&d,1:dopts[&d,5]]],nrow(evt),1);
          end;
/* dopts[,4] = 0 means don't transform the tie-list so do nothing here */
          dvec2 = dvec` * evt_i;
          dline = dvec2[+];
          if (dopts[&d,4] = 1) & (dopts[&d,2] = 2) then do;
                if any(diffx[i,dvars] ^= mvalue) then do;
                   dvars_i = diffx[i,dvars];
                   dvars_i = dvars_i[loc(dvars_i ^= mvalue)];
                   wait_i = nrow(dvars_i);
                  if (wait_i ^= 0) then dline = dline/wait_i;
                end;
          end;
          else if (dopts[&d,4] = 1) & (dopts[&d,2] = 1) then do; 
                   links_i = (x[,dvars] = diffx[i,id]);
                   wait_i = sum(links_i);
                   if (wait_i ^= 0) then dline = dline/wait_i;
          end;    
       end;

       if (&d = 1) then
          covd_s = dline;
       else
          covd_s = covd_s || dline;

      %end;
    end;

    if (bterm ^= 0) | (cterm ^= 0) then do;
         diffline = covb_s || covc_s;
         if (dterm ^= 0) then diffline = diffline || covd_s;
    end;
    else if (dterm ^= 0) then diffline = covd_s;

      
    dcovmx = dcovmx // diffline;

end;


if dterm ^= 0 then do;
    do dp = 1 to nrow(dopts);
       if ((dopts[dp,1] = 1) & (dopts[dp,2] = 4)) then do;
             dcovmx[,d_strt - dif_strt + dp] 
               = (max_d[dp] * nevts) - dcovmx[,d_strt - dif_strt + dp];
       end;
    end;
end;

    print "Descriptive Statistics: Diffusion Effects";
    run dstats(dcovmx,dcovname); 


/* add diffusion covariates to diffx */
diffx = diffx || dcovmx;

free dcovmx evt_i;


/* set up optimization routines */

         diffvec = ncol(diffx) - nrow(covnames) + dif_strt : ncol(diffx);
         if (aterm = 1) then 
             covx = diffvec;
         else if (aterm = 2)|(aterm = 3) then
             covx = avars || diffvec;
         dxnames = xvars // dcovname;
         print "Descriptive Statistics: Diffusion Matrix";
         run dstats(diffx,dxnames);

end;
else do;
    diffx = x;
    covx = avars;

    run dstats(diffx,xvars);
end;

if ((verbose = 1) & (use_et2 = 1)) then do;
    ts_x = diffx[,id] || diffx[,st] || diffx[,et] || 
           diffx[,es] || diffx[,et2] || diffx[,es2]; 
    print "Diffusion spells" ts_x;
    ts_c = diffx[,id] || diffx[,st] || diffx[,et] || 
           diffx[,diffvec];
    print "Diffusion covariates" ts_c;
end;


if (save_all = 1) then do;
     allnames = xvars // dcovname;
     create fromdiff from diffx [colname = allnames]; 
     append from diffx;
     close fromdiff;
end;

         xcovx = diffx[,covx];
         ncases = nrow(diffx);
         ncovs =  ncol(xcovx);

         risktime = diffx[,et] - diffx[,st];
         timet = sum(risktime);
         if (xtimet ^= timet) then
print "Warning: diffx time at risk does not equal original" xtimet timet;

         events   = diffx[,es] ^= diffx[,ss];
         nevents = sum(events);
         if (xnevents ^= nevents) then
print "Warning: diffx events does not equal original" xnevents nevents;


print "Diffusion Analysis";
print "N Spells " ncases "Total Time at Risk " timet "N Events " nevents;

         x0 = repeat(1,ncases,1);
         xcovx = x0 || xcovx;

         b0 = log(nevents/timet);
         b_init = b0 || repeat(0,1,ncovs);
         if (&user_est = 1) then do;
              b_init_0 = b_init;
              b_init = {&init_est};
         end;


/*
dmx = diffx[,id] || diffx[,st] || diffx[,et] || 
           diffx[,es] || diffx[,covx];
dmv = xvars[id] || xvars[st] || xvars[et] || 
           xvars[es] ||  cov_nint`;
if (wrtvars ^= 0) then do;
     dmx = dmx || diffx[,wrtvars];
     dmv = dmv ||  xvars[wrtvars];
end;


create diffset from dmx [colname = dmv]; 
append from dmx;
close diffset;


free dmx dmv;
*/
free diffx;

/* -----  likelihood functions ----- */

   evx = events` * xcovx;

start f_expon(b) global(evx,risktime,xcovx,rexpbx);
   rexpbx = risktime # exp(xcovx * b`);
   sumLL = sum(evx # b) - sum(rexpbx);
   return(sumLL);
finish f_expon;

start g_expon(b) global(evx,risktime,xcovx,rexpbx);
    g_exp = evx - rexpbx` * xcovx;
    return(g_exp);
finish g_expon;

start h_expon(b) global(evx,risktime,xcovx,rexpbx);
    h_exp = - (xcovx # rexpbx)` * xcovx;
    return(h_exp);
finish h_expon;


opt = {1 2};
call NLPNRA(rc,xres,"f_expon",b_init,opt) grd="g_expon" hes="h_expon";


/* calculate goodness of fit statistics */
if (&user_est = 1) then b_init = b_init_0;
LL_0 = f_expon(b_init);
LL_diff = f_expon(xres);
lr_0 = 2 * (LL_diff - LL_0);
p_lr_0 = 1 - probchi(lr_0,ncol(b_init) - 1);

/* calculate significance levels */
final_g = g_expon(xres);
/* don't work w gg (parameter var-covar matrix) in this version;
   investigate in mhdiff22.sas 
gg = final_g` * final_g;
gginv = inv(gg); */
final_h = h_expon(xres);
covest = inv(-final_h);
stderr = sqrt(vecdiag(covest));
xres = xres`;
cferr = xres/stderr;
probhat = 2 # (1 - probnorm(abs(cferr)));


/* print results */

results = xres || stderr || cferr || probhat;
rcols = {"Est'd B" "S.E." "B/S.E." "Pr(B = 0)"};

print "Multiplicative diffusion model results: ";

print results [r=covnames c = rcols format=10.6];

print "Log Likelihood: "  LL_diff,
      "Likelihood ratio vs r(t) = a_0: "        lr_0,
      "Chi-square prob level for H_0 = H_alt: "  p_lr_0;


/* exit iml */
quit;

/* close macro */
%mend;

