/* 
This file illustrates an invocation of mhdiff, using data from Coleman et al/Burt's study of tetracycline adoption. Run "Coleman_data_setup.sas" first to generate the SAS datasets referred to here 
*/

/* ------- multiplicative heterogeneous diffusion model ---------- */
/*                                                                 */
/*       example program :                                         */
/*            illustrates how to setup and invoke mhdiff           */
/*                                                                 */
/*       David Strang/ November 1995                               */
/* --------------------------------------------------------------- */


options ls = 72;


/* ------------------- */
/* external data setup */
/* ------------------- */


libname datadir '.';

%let indir = 0;   /* = 0 if data in SAS working datasets, 
                     = 1 if in external files */

data cole_spl;
     set datadir.cole1;

if city = 3;

data nxnd_3;
      set datadir.cprx;

/* cprx holds a full influence matrix */


/* ------------------------ */
/* define spell identifiers */
/* ------------------------ */


%let sdata = cole_spl;   /* input spell data set */

%let sort_var = emonth;    /* name of end time variable in sdata */

%let st = 4;             /* start time variable */

%let et = 5;             /* end time variable */

%let ss = 2;             /* start state variable */

%let es = 3;             /* end time variable */

%let id = 1;             /* numeric case id */

%let mvalue = 999;       /* missing value coding for id */



/* ------------ */
/* define model */
/* ------------ */

%let aterm = 3;                 /* propensity */
%let avars = 13 14 15;          

%let bterm = 3;                 /* susceptibility */
%let bvars = 12;

%let cterm = 1;                 /* contagiousness */
%let cvars = 12;

%let dterm = 3;                 /* social proximity */

                /* term 1 : social cohesion via tie-list                 */
                /* term 2 : same "science" orientation via power metric  */
                /* term 3 : structural equivalence via full matrix       */

%let dopts =  3 2 1 0 6,
              1 2 0 1 1,
              2 1 0 0 1;

%let dvecs =   6  7  8  9 10 11,
              13  0  0  0  0  0,
               0  0  0  0  0  0;

%let dnames = "cohesion", "med. orient", "struc. equiv";




%let use_et2 = 0;

%let et2 = 11;

%let es2 = 11;

%let st2 = 11;

%let ss2 = 11;


/* ------------------ */
/* additional options */
/* ------------------ */

%let save_all = 1;

%let wspace = 512;  /* size of workspace to be allocated */

%let verbose = 1;  /* print additional statistics + descriptive info */

%let sametime = 1;  /* 0 if simultaneous influence disallowed, 1 if allowed */

%let simlapse = 0.5; 
                    /* lapse period over which simultaneous influence occurs */

%let insum = 1;     /* 0 if all are potential influencers; 1 if influencer
                       must be in same "igroup" as influencee */

%let igroup = 18;   /* location of igroup variable (here, city) */


%let user_est = 0;  /* user provides initial estimates (here, no) */

%let init_est = -4.07 0.42 -.96 .63 .0198 2.34 0.0 -6 .23 9.99;
                    /* example of user provided estimates vector */

%let tiesize = 0;   /* strength of ties; not used in this example */




/* ----------------- */
/* invoke mhdiff.sas */
/* ----------------- */

%include mhdiff_b/source2;
%mhdiffb2;

